@@ -24,9 +24,10 @@ CHANGES:
 As of 0.3.0, options should be specified as `optparse` options.  See README.txt.
 flagReader.py options are still supported for backward compatibility
 """
-import cmd, re, os, sys, optparse, subprocess, tempfile, pyparsing, doctest, unittest, string
+import cmd, re, os, sys, optparse, subprocess, tempfile, pyparsing, doctest
+import unittest, string, datetime
 from optparse import make_option
-__version__ = '0.4.5'
+__version__ = '0.4.7'
 
 class OptionParser(optparse.OptionParser):
     def exit(self, status=0, msg=None):
@@ -48,7 +49,7 @@ def remainingArgs(oldArgs, newArgList):
     >>> remainingArgs('-f bar   bar   cow', ['bar', 'cow'])
     'bar   cow'
     '''
-    pattern = '\s+'.join(newArgList) + '\s*$'
+    pattern = '\s+'.join(re.escape(a) for a in newArgList) + '\s*$'
     matchObj = re.search(pattern, oldArgs)
     return oldArgs[matchObj.start():]
 
@@ -60,8 +61,8 @@ def options(option_list):
         optionParser.set_usage("%s [options] arg" % func.__name__.strip('do_'))
         def newFunc(instance, arg):
             try:
-                opts, newArgList = optionParser.parse_args(arg.split())
-                newArgs = remainingArgs(arg, newArgList)
+                opts, newArgList = optionParser.parse_args(arg.split()) # doesn't understand quoted strings shouldn't be dissected!
+                newArgs = remainingArgs(arg, newArgList)  # should it permit flags after args?
             except (optparse.OptionValueError, optparse.BadOptionError,
                     optparse.OptionError, optparse.AmbiguousOptionError,
                     optparse.OptionConflictError), e:
@@ -153,51 +154,75 @@ else:
         writeToPasteBuffer = getPasteBuffer
           
 pyparsing.ParserElement.setDefaultWhitespaceChars(' \t')
-'''
-def parseSearchResults(pattern, s):
-    generator = pattern.scanString(s)
-    try:
-        result, start, stop = generator.next()
-        result['before'], result['after'] = s[:start], s[stop:]
-        result['upToIncluding'] = s[:stop]
-    except StopIteration:
-        result = pyparsing.ParseResults('')
-        result['before'] = s
-    return result
-'''
-
-def replaceInput(source):
-    if source:
-        newinput = open(source[0], 'r').read()
-    else:
-        newinput = getPasteBuffer()
-    
-        try:
-            if statement.inputFrom:
-                newinput = open(statement.inputFrom, 'r').read()
-            else:
-                newinput = getPasteBuffer()
-        except (OSError,), e:
-            print e
-            return 0                    
-        start, end = self.redirectInPattern.scanString(statement.fullStatement).next()[1:]
-        return self.onecmd('%s%s%s' % (statement.fullStatement[:start], 
-                            newinput, statement.fullStatement[end:]))
 
 class ParsedString(str):
     pass
 
+class SkipToLast(pyparsing.SkipTo):
+    def parseImpl( self, instring, loc, doActions=True ):
+        resultStore = []
+        startLoc = loc
+        instrlen = len(instring)
+        expr = self.expr
+        failParse = False
+        while loc <= instrlen:
+            try:
+                if self.failOn:
+                    failParse = True
+                    self.failOn.tryParse(instring, loc)
+                    failParse = False
+                loc = expr._skipIgnorables( instring, loc )
+                expr._parse( instring, loc, doActions=False, callPreParse=False )
+                skipText = instring[startLoc:loc]
+                if self.includeMatch:
+                    loc,mat = expr._parse(instring,loc,doActions,callPreParse=False)
+                    if mat:
+                        skipRes = ParseResults( skipText )
+                        skipRes += mat
+                        resultStore.append((loc, [ skipRes ]))
+                    else:
+                        resultStore,append((loc, [ skipText ]))
+                else:
+                    resultStore.append((loc, [ skipText ]))
+                loc += 1
+            except (pyparsing.ParseException,IndexError):
+                if failParse:
+                    raise
+                else:
+                    loc += 1
+        if resultStore:
+            return resultStore[-1]
+        else:
+            exc = self.myException
+            exc.loc = loc
+            exc.pstr = instring
+            raise exc    
+
+def replace_with_file_contents(fname):
+    import pdb; pdb.set_trace()
+    if fname:
+        try:
+            result = open(os.path.expanduser(fname[0])).read()
+        except IOError:
+            result = '< %s' % fname[0]  # wasn't a file after all
+    else:
+        result = getPasteBuffer()
+    return result
+        
 class Cmd(cmd.Cmd):
     echo = False
     caseInsensitive = True
     continuationPrompt = '> '  
+    timing = False
     legalChars = '!#$%.:?@_' + pyparsing.alphanums + pyparsing.alphas8bit  # make sure your terminators are not in here!
     shortcuts = {'?': 'help', '!': 'shell', '@': 'load' }
     excludeFromHistory = '''run r list l history hi ed edit li eof'''.split()
     noSpecialParse = 'set ed edit exit'.split()
     defaultExtension = 'txt'
     defaultFileName = 'command.txt'
-    settable = ['prompt', 'continuationPrompt', 'defaultFileName', 'editor', 'caseInsensitive', 'echo']
+    settable = ['prompt', 'continuationPrompt', 'defaultFileName', 'editor', 'caseInsensitive', 
+                'echo', 'timing']
+    settable.sort()
     
     editor = os.environ.get('EDITOR')
     _STOP_AND_EXIT = 2
@@ -210,12 +235,13 @@ class Cmd(cmd.Cmd):
                     break
             
     def do_cmdenvironment(self, args):
+        '''Summary report of interactive parameters.'''
         self.stdout.write("""
         Commands are %(casesensitive)scase-sensitive.
         Commands may be terminated with: %(terminators)s
         Settable parameters: %(settable)s
         """ % 
-        { 'casesensitive': ('not ' and self.caseInsensitive) or '',
+        { 'casesensitive': (self.caseInsensitive and 'not ') or '',
           'terminators': str(self.terminators),
           'settable': ' '.join(self.settable)
         })
@@ -304,6 +330,13 @@ class Cmd(cmd.Cmd):
         - pipeTo:  piped
         - statement: ['simple', '']
           - command: simple
+        >>> print c.parser.parseString('double-pipe || is not a pipe').dump()
+        ['double', '-pipe || is not a pipe']
+        - args: -pipe || is not a pipe
+        - command: double
+        - statement: ['double', '-pipe || is not a pipe']
+          - args: -pipe || is not a pipe
+          - command: double
         >>> print c.parser.parseString('command with args, terminator;sufx | piped').dump()
         ['command', 'with args, terminator', ';', 'sufx', '|', ' piped']
         - args: with args, terminator
@@ -397,14 +430,15 @@ class Cmd(cmd.Cmd):
           - terminator: ['\n', '\n']
         - terminator: ['\n', '\n']
         '''
-        outputParser = pyparsing.oneOf(['>>','>'])('output')
+        outputParser = (pyparsing.Literal('>>') | (pyparsing.WordStart() + '>') | pyparsing.Regex('[^=]>'))('output')
+        
         terminatorParser = pyparsing.Or([(hasattr(t, 'parseString') and t) or pyparsing.Literal(t) for t in self.terminators])('terminator')
         stringEnd = pyparsing.stringEnd ^ '\nEOF'
         self.multilineCommand = pyparsing.Or([pyparsing.Keyword(c, caseless=self.caseInsensitive) for c in self.multilineCommands])('multilineCommand')
         oneLineCommand = (~self.multilineCommand + pyparsing.Word(self.legalChars))('command')
         pipe = pyparsing.Keyword('|', identChars='|')
-        self.commentGrammars.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).setParseAction(lambda x: '')
-        self.commentInProgress.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(pyparsing.cStyleComment)       
+        self.commentGrammars.ignore(pyparsing.quotedString).setParseAction(lambda x: '')
+        self.commentInProgress.ignore(pyparsing.quotedString).ignore(pyparsing.cStyleComment)       
         afterElements = \
             pyparsing.Optional(pipe + pyparsing.SkipTo(outputParser ^ stringEnd)('pipeTo')) + \
             pyparsing.Optional(outputParser + pyparsing.SkipTo(stringEnd).setParseAction(lambda x: x[0].strip())('outputTo'))
@@ -417,7 +451,7 @@ class Cmd(cmd.Cmd):
             self.blankLineTerminator = (pyparsing.lineEnd + pyparsing.lineEnd)('terminator')
             self.blankLineTerminator.setResultsName('terminator')
             self.blankLineTerminationParser = ((self.multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(self.blankLineTerminator).setParseAction(lambda x: x[0].strip())('args') + self.blankLineTerminator)('statement')
-        self.multilineParser = (((self.multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(terminatorParser).setParseAction(lambda x: x[0].strip())('args') + terminatorParser)('statement') +
+        self.multilineParser = (((self.multilineCommand ^ oneLineCommand) + SkipToLast(terminatorParser).setParseAction(lambda x: x[0].strip())('args') + terminatorParser)('statement') +
                                 pyparsing.SkipTo(outputParser ^ pipe ^ stringEnd).setParseAction(lambda x: x[0].strip())('suffix') + afterElements)
         self.singleLineParser = ((oneLineCommand + pyparsing.SkipTo(terminatorParser ^ stringEnd ^ pipe ^ outputParser).setParseAction(lambda x:x[0].strip())('args'))('statement') +
                                  pyparsing.Optional(terminatorParser) + afterElements)
@@ -431,14 +465,18 @@ class Cmd(cmd.Cmd):
             self.blankLineTerminationParser | 
             self.multilineCommand + pyparsing.SkipTo(stringEnd)
             )
-        self.parser.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)
+        self.parser.ignore(pyparsing.quotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)
         
         inputMark = pyparsing.Literal('<')
         inputMark.setParseAction(lambda x: '')
-        inputFrom = pyparsing.Word(self.legalChars + '/\\')('inputFrom')
-        inputFrom.setParseAction(lambda x: (x and open(x[0]).read()) or getPasteBuffer())
-        self.inputParser = inputMark + pyparsing.Optional(inputFrom)
-        self.inputParser.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)               
+        fileName = pyparsing.Word(self.legalChars + '/\\')
+        inputFrom = fileName('inputFrom')
+        inputFrom.setParseAction(replace_with_file_contents)
+        # a not-entirely-satisfactory way of distinguishing < as in "import from" from <
+        # as in "lesser than"
+        self.inputParser = inputMark + pyparsing.Optional(inputFrom) + pyparsing.Optional('>') + \
+                           pyparsing.Optional(fileName) + (pyparsing.stringEnd | '|')
+        self.inputParser.ignore(pyparsing.quotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)               
     
     def parsed(self, raw, **kwargs):
         if isinstance(raw, ParsedString):
@@ -446,7 +484,7 @@ class Cmd(cmd.Cmd):
         else:
             s = self.inputParser.transformString(raw.lstrip())
             for (shortcut, expansion) in self.shortcuts.items():
-                if s.startswith(shortcut):
+                if s.lower().startswith(shortcut):
                     s = s.replace(shortcut, expansion + ' ', 1)
                     break
             result = self.parser.parseString(s)
@@ -480,8 +518,9 @@ class Cmd(cmd.Cmd):
         try:
             statement = self.parsed(line)
             while statement.parsed.multilineCommand and (statement.parsed.terminator == ''):
-                statement = self.parsed('%s\n%s' % (statement.parsed.raw, 
-                                        self.pseudo_raw_input(self.continuationPrompt)))
+                statement = '%s\n%s' % (statement.parsed.raw, 
+                                        self.pseudo_raw_input(self.continuationPrompt))                
+                statement = self.parsed(statement)
         except Exception, e:
             print e
             return 0
@@ -503,7 +542,7 @@ class Cmd(cmd.Cmd):
                 if statement.parsed.output == '>>':
                     mode = 'a'
                 try:
-                    self.stdout = open(statement.parsed.outputTo, mode)                            
+                    self.stdout = open(os.path.expanduser(statement.parsed.outputTo), mode)                            
                 except OSError, e:
                     print e
                     return 0                    
@@ -519,7 +558,10 @@ class Cmd(cmd.Cmd):
                 func = getattr(self, 'do_' + statement.parsed.command)
             except AttributeError:
                 return self.default(statement)
-            stop = func(statement)                        
+            timestart = datetime.datetime.now()
+            stop = func(statement) 
+            if self.timing:
+                print 'Elapsed: %s' % str(datetime.datetime.now() - timestart)
         except Exception, e:
             print e
         try:
@@ -610,10 +652,11 @@ class Cmd(cmd.Cmd):
                
     def showParam(self, param):
         param = param.strip().lower()
-        if param in self.settable:
-            val = getattr(self, param)
-            self.stdout.write('%s: %s\n' % (param, str(getattr(self, param))))
-
+        for p in self.settable:
+            if p.startswith(param):
+                val = getattr(self, p)
+                self.stdout.write('%s: %s\n' % (p, str(getattr(self, p))))
+                
     def do_quit(self, arg):
         return self._STOP_AND_EXIT
     do_exit = do_quit
@@ -628,22 +671,25 @@ class Cmd(cmd.Cmd):
                 self.showParam(param)
     
     def do_set(self, arg):
-        'Sets a parameter'        
+        '''Sets a cmd2 parameter.  Accepts abbreviated parameter names so long as there is no ambiguity.
+           Call without arguments for a list of settable parameters with their values.'''
         try:
             paramName, val = arg.split(None, 1)
             paramName = paramName.strip().lower()
-            if paramName not in self.settable:
-                raise NotSettableError
-            currentVal = getattr(self, paramName)
-            if (val[0] == val[-1]) and val[0] in ("'", '"'):
-                val = val[1:-1]
-            else:                
-                val = cast(currentVal, val)
-            setattr(self, paramName, val)
-            self.stdout.write('%s - was: %s\nnow: %s\n' % (paramName, currentVal, val))
+            hits = [paramName in p for p in self.settable]
+            if hits.count(True) == 1:
+                paramName = self.settable[hits.index(True)]
+                currentVal = getattr(self, paramName)
+                if (val[0] == val[-1]) and val[0] in ("'", '"'):
+                    val = val[1:-1]
+                else:                
+                    val = cast(currentVal, val)
+                setattr(self, paramName, val)
+                self.stdout.write('%s - was: %s\nnow: %s\n' % (paramName, currentVal, val))
+            else:
+                self.do_show(paramName)
         except (ValueError, AttributeError, NotSettableError), e:
             self.do_show(arg)
-    do_set.__doc__ = '%s\nOne of: %s' % (do_set.__doc__, ', '.join(settable))
                 
     def do_pause(self, arg):
         'Displays the specified text then waits for the user to press RETURN.'
@@ -714,7 +760,7 @@ class Cmd(cmd.Cmd):
             buffer = self.history[-1]
 
         if buffer:
-            f = open(filename, 'w')
+            f = open(os.path.expanduser(filename), 'w')
             f.write(buffer or '')
             f.close()        
                 
@@ -744,7 +790,7 @@ class Cmd(cmd.Cmd):
         else:
             saveme = self.history[-1]
         try:
-            f = open(fname, 'w')
+            f = open(os.path.expanduser(fname), 'w')
             f.write(saveme)
             f.close()
             print 'Saved to %s' % (fname)
@@ -754,16 +800,17 @@ class Cmd(cmd.Cmd):
     def do_load(self, fname=None):
         """Runs command(s) from a file."""
         if fname is None:
-            fname = self.defaultFileName        
+            fname = self.defaultFileName
+        fname = os.path.expanduser(fname)
         keepstate = Statekeeper(self, ('stdin','use_rawinput','prompt','continuationPrompt'))
         if isinstance(fname, file):
             self.stdin = fname
         else:           
             try:
-                self.stdin = open(fname, 'r')
+                self.stdin = open(os.path.expanduser(fname), 'r')
             except IOError, e:
                 try:
-                    self.stdin = open('%s.%s' % (fname, self.defaultExtension), 'r')
+                    self.stdin = open('%s.%s' % (os.path.expanduser(fname), self.defaultExtension), 'r')
                 except IOError:
                     print 'Problem opening file %s: \n%s' % (fname, e)
                     keepstate.restore()
@@ -796,7 +843,7 @@ class Cmd(cmd.Cmd):
             
     def fileimport(self, statement, source):
         try:
-            f = open(source)
+            f = open(os.path.expanduser(source))
         except IOError:
             self.stdout.write("Couldn't read from file %s\n" % source)
             return ''
@@ -918,7 +965,7 @@ class Cmd2TestCase(unittest.TestCase):
             self.outputTrap = OutputTrap()
             self.cmdapp = self.CmdApp()
             try:
-                tfile = open(self.transcriptFileName)
+                tfile = open(os.path.expanduser(self.transcriptFileName))
                 self.transcript = iter(tfile.readlines())
                 tfile.close()
             except IOError:
